/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"


#include "ultra_r4300.h"

#include "DBGConsole.h"

#include "Registers.h"	// For RegNames
#include "CPU.h"
#include "RSP.h"
#include "RDP.h"
#include "RDP_GFX.h"
#include "OpCode.h"
#include "Patch.h"		// To get patch names
#include "Dump.h"
#include "Controller.h"

// Size of the background text buffer
static const COORD sc_outbufsize = { 80, 120 };

static HANDLE g_hStdOut = NULL;
static HANDLE g_hOutputBuffer = INVALID_HANDLE_VALUE;
static HANDLE g_hConsoleThread = NULL;

static BOOL DBGConsole_Init();
static void DBGConsole_Fini();

static BOOL DBGConsole_ResizeConBufAndWindow(HANDLE hConsole, SHORT xSize, SHORT ySize);
static DWORD __stdcall DBGConsole_ConsoleFunc(LPVOID pArg);
static void DBGConsole_ProcessKeyEvent(KEY_EVENT_RECORD kr);

static void DBGConsole_ParseStringHighlights(LPSTR szString, WORD * arrwAttributes, WORD wAttr);
static BOOL DBGConsole_WriteString(LPCTSTR szString, BOOL bParse, SHORT x, SHORT y, WORD wAttr, SHORT width);
static void DBGConsole_DisplayOutput(LONG nOffset);

enum DBGCON_MIDPANETYPE
{
	DBGCON_CPU_CODE_VIEW = 0,
	DBGCON_RSP_CODE_VIEW,
};

CHAR * s_szPaneNameLabels[] = { "Disassembly", "RSP Disassembly", "Memory" };

enum DBGCON_PANETYPE
{
	DBGCON_REG_PANE = 0,		// Register view
	DBGCON_MEM_PANE,
	DBGCON_MID_PANE,			// Disasm/Mem area
	DBGCON_BOT_PANE				// Output area
};

static DWORD g_dwMemAddress = 0x80000000;
static DWORD g_dwDisplayPC = ~0;
static DWORD g_dwDisplayRSP_PC = ~0;
static LONG  g_nOutputBufOffset = 0;

static DBGCON_MIDPANETYPE s_nMidPaneType = DBGCON_CPU_CODE_VIEW;

static DBGCON_PANETYPE s_nActivePane = DBGCON_MID_PANE;

static void DBGConsole_SetActivePane(DBGCON_PANETYPE pane);
static void DBGConsole_RedrawPaneLabels();

static void DBGConsole_ProcessInput();
static void DBGConsole_HelpCommand();
static void DBGConsole_GoCommand();
static void DBGConsole_StopCommand();

//////////////////////////////////////////////////////////////////////////
static void DBGConsole_DumpContext(DWORD dwPC);
static void DBGConsole_DumpContext_NextLine();
static void DBGConsole_DumpContext_NextPage();
static void DBGConsole_DumpContext_PrevLine();
static void DBGConsole_DumpContext_PrevPage();

static void DBGConsole_DumpMemory(DWORD dwAddress);
static void DBGConsole_DumpCPUContext(DWORD dwPC);
static void DBGConsole_DumpRSPContext(DWORD dwPC);


//////////////////////////////////////////////////////////////////////////
static void DBGConsole_DumpRegisters();
static void DBGConsole_DumpCPURegisters();
static void DBGConsole_DumpRSPRegisters();



//////////////////////////////////////////////////////////////////////////
static const WORD sc_wAttrWhite = FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE;
static const WORD sc_wAttrBoldWhite = FOREGROUND_INTENSITY|FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE;
static const WORD sc_wAttrRed = FOREGROUND_RED;
static const WORD sc_wAttrBoldRed = FOREGROUND_INTENSITY|FOREGROUND_RED;

static TCHAR g_szInput[1024+1];

static const int sc_nTopPaneTextOffset = 0;  static const sc_nTopPaneLines = 11;
static const int sc_nMemPaneTextOffset = 12; static const sc_nMemPaneLines = 7;
static const int sc_nMidPaneTextOffset = 20; static const sc_nMidPaneLines = 13;
static const int sc_nBotPaneTextOffset = 34; static const sc_nBotPaneLines = 24;
static const int sc_nCommandTextOffset = 59 ;



typedef struct tagDebugCommandInfo
{
	LPCTSTR szCommand;
	void (* pArglessFunc)();
	void (* pArgFunc)(LPCTSTR);
} DebugCommandInfo;

static void DBGConsole_IntPi() { DBGConsole_Msg(0, "Pi Interrupt"); Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_PI); AddCPUJob(CPU_CHECK_INTERRUPTS); }
static void DBGConsole_IntAi() { DBGConsole_Msg(0, "Ai Interrupt"); Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_AI); AddCPUJob(CPU_CHECK_INTERRUPTS); }
static void DBGConsole_IntDp() { DBGConsole_Msg(0, "Dp Interrupt"); Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_DP); AddCPUJob(CPU_CHECK_INTERRUPTS); }
static void DBGConsole_IntSp() { DBGConsole_Msg(0, "Sp Interrupt"); Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_SP); AddCPUJob(CPU_CHECK_INTERRUPTS); }
static void DBGConsole_IntSi() { DBGConsole_Msg(0, "Si Interrupt"); Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_SI); AddCPUJob(CPU_CHECK_INTERRUPTS); }
static void DBGCommand_Mem(LPCTSTR szCommand);
static void DBGCommand_List(LPCTSTR szCommand);
static void DBGCommand_ListOS(LPCTSTR szCommand);
static void DBGCommand_BPX(LPCTSTR szCommand);
static void DBGCommand_ShowCPU();
static void DBGCommand_ShowRSP();


#define BEGIN_DBGCOMMAND_MAP(name) \
static const DebugCommandInfo name[] =	\
{
#define DBGCOMMAND_ARGLESS(cmd, func) \
{ cmd, func, NULL },
#define DBGCOMMAND_ARG(cmd, func) \
{ cmd, NULL, func },

#define END_DBGCOMMAND_MAP()		\
	{ NULL, NULL }					\
};



BEGIN_DBGCOMMAND_MAP(g_DBGCommands)
	DBGCOMMAND_ARGLESS( "help", DBGConsole_HelpCommand )
	DBGCOMMAND_ARGLESS( "go", DBGConsole_GoCommand )
	DBGCOMMAND_ARGLESS( "stop", DBGConsole_StopCommand )
	DBGCOMMAND_ARGLESS( "skip", CPUSkip )
	DBGCOMMAND_ARGLESS( "dump dlist", RDP_GFX_DumpNextDisplayList )
	DBGCOMMAND_ARGLESS( "dump textures", RDP_GFX_DumpTextures )
	DBGCOMMAND_ARGLESS( "nodump textures", RDP_GFX_NoDumpTextures )
	DBGCOMMAND_ARGLESS( "drop textures", RDP_GFX_DropTextures )
	DBGCOMMAND_ARGLESS( "make textures blue", RDP_GFX_MakeTexturesBlue )
	DBGCOMMAND_ARGLESS( "make textures normal", RDP_GFX_MakeTexturesNormal )
	DBGCOMMAND_ARGLESS( "enablegfx", RDP_EnableGfx )
	DBGCOMMAND_ARGLESS( "disablegfx", RDP_DisableGfx )
	
	DBGCOMMAND_ARGLESS( "stat dr", SR_Stats )

	DBGCOMMAND_ARGLESS( "disable si", Memory_SiDisable )
	DBGCOMMAND_ARGLESS( "enable si", Memory_SiEnable )

	DBGCOMMAND_ARGLESS( "int pi", DBGConsole_IntPi )
	DBGCOMMAND_ARGLESS( "int ai", DBGConsole_IntAi )
	DBGCOMMAND_ARGLESS( "int dp", DBGConsole_IntDp )
	DBGCOMMAND_ARGLESS( "int sp", DBGConsole_IntSp )
	DBGCOMMAND_ARGLESS( "int si", DBGConsole_IntSi )

	DBGCOMMAND_ARG( "dump bootcode", Dump_BootROMRange )
	DBGCOMMAND_ARG( "dump os", Dump_OSRange )
	DBGCOMMAND_ARG( "dump exvec", Dump_ExceptionVectors )
		
	DBGCOMMAND_ARG( "listos", DBGCommand_ListOS )
	DBGCOMMAND_ARG( "list", DBGCommand_List )
	DBGCOMMAND_ARGLESS( "cpu", DBGCommand_ShowCPU )
	DBGCOMMAND_ARGLESS( "rsp", DBGCommand_ShowRSP )
	DBGCOMMAND_ARG( "mem", DBGCommand_Mem )
	DBGCOMMAND_ARG( "bpx", DBGCommand_BPX )

	DBGCOMMAND_ARGLESS( "osinfo threads", Patch_DumpOsThreadInfo )
	DBGCOMMAND_ARGLESS( "osinfo queues", Patch_DumpOsQueueInfo )
	DBGCOMMAND_ARGLESS( "osinfo events", Patch_DumpOsEventInfo )
	
END_DBGCOMMAND_MAP()

static const DWORD sc_Cop0RegsToShow[] =
{
	C0_INX,
	//C0_RAND,
	C0_ENTRYLO0,
	C0_ENTRYLO1,
	C0_CONTEXT,
	C0_PAGEMASK,
	C0_WIRED,
	C0_BADVADDR,
	C0_COUNT,
	C0_ENTRYHI,
	C0_SR,
	C0_CAUSE,
	C0_EPC,
	//C0_PRID,
	C0_COMPARE,
	C0_CONFIG,
	C0_LLADDR,
	C0_WATCHLO,
	C0_WATCHHI,
	C0_ECC,			// PERR
	C0_CACHE_ERR,
	C0_TAGLO,
	C0_TAGHI,
	C0_ERROR_EPC,
	~0
};

static BOOL s_bConsoleAllocated = FALSE;

void DBGConsole_Enable(BOOL bEnable)
{
	// Shut down
	DBGConsole_Fini();

	if (bEnable)
	{
		g_bShowDebug = DBGConsole_Init();
	}
	else
	{
		g_bShowDebug = FALSE;
	}

}


BOOL DBGConsole_Init()
{
	BOOL bRetVal;

	// Don't re-init!
	if (s_bConsoleAllocated)
		return TRUE;
	
	bRetVal = AllocConsole();
	if (!bRetVal)
		return FALSE;

	s_bConsoleAllocated = TRUE;
 
    // Get a handle to the STDOUT screen buffer to copy from and 
    // create a new screen buffer to copy to. 
 
    g_hStdOut = GetStdHandle(STD_OUTPUT_HANDLE); 
    if (g_hStdOut == INVALID_HANDLE_VALUE) 
        return FALSE;
 
    g_hOutputBuffer = CreateConsoleScreenBuffer( 
       GENERIC_READ |           // read-write access 
       GENERIC_WRITE, 
       0,                       // not shared 
       NULL,                    // no security attributes 
       CONSOLE_TEXTMODE_BUFFER, // must be TEXTMODE 
       NULL);                   // reserved; must be NULL 
    if (g_hOutputBuffer == INVALID_HANDLE_VALUE) 
        return FALSE;

	// Resize to some reasonable size
	if (!SetConsoleScreenBufferSize(g_hOutputBuffer, sc_outbufsize))
		return FALSE;

	DBGConsole_ResizeConBufAndWindow(g_hStdOut, 80,60);

	DBGConsole_SetActivePane(DBGCON_BOT_PANE);

	DBGConsole_DumpRegisters();
	DBGConsole_DumpMemory(~0);
	DBGConsole_DumpContext(~0);


	DWORD dwThreadID;

	g_hConsoleThread = CreateThread(NULL, 0, DBGConsole_ConsoleFunc, NULL, CREATE_SUSPENDED , &dwThreadID);
	if (g_hConsoleThread == NULL)
		return FALSE;

	ResumeThread(g_hConsoleThread);

	return TRUE;
}

void DBGConsole_Fini()
{
	if (g_hConsoleThread != NULL)
	{
		CloseHandle(g_hConsoleThread);
		g_hConsoleThread = NULL;
	}

	// Maybe preserve this - we might want to use the contents later
	if (g_hOutputBuffer != INVALID_HANDLE_VALUE)
	{
		CloseHandle(g_hOutputBuffer);
		g_hOutputBuffer = INVALID_HANDLE_VALUE;
	}

	if (s_bConsoleAllocated)
	{
		FreeConsole();
		s_bConsoleAllocated = FALSE;
	}
}

void DBGConsole_UpdateDisplay()
{
	// Return if we don't have a console!
	if (!g_bShowDebug || !s_bConsoleAllocated)
		return;

	DBGConsole_DumpRegisters();
	DBGConsole_DumpMemory(~0);
	DBGConsole_DumpContext(~0);
}


DWORD __stdcall DBGConsole_ConsoleFunc(LPVOID pArg)
{
    DWORD cNumRead, fdwMode, fdwSaveOldMode, i; 
    INPUT_RECORD irInBuf[128]; 
	COORD curPos;

    HANDLE hInput = GetStdHandle(STD_INPUT_HANDLE); 
    if (hInput == INVALID_HANDLE_VALUE) 
        return 1;


    // Save the current input mode, to be restored on exit. 
    if (!GetConsoleMode(hInput, &fdwSaveOldMode) ) 
        return 1; 
 
    // Enable the window and mouse input events. 
    fdwMode = ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT; 
    if (! SetConsoleMode(hInput, fdwMode) ) 
        return 1; 

	curPos.X = 0;
	curPos.Y = sc_nCommandTextOffset;
	SetConsoleCursorPosition(g_hStdOut, curPos);
 	

	// Handle set to null on exit
	while (g_hConsoleThread != NULL)
	{
		
        if (! ReadConsoleInput( 
                hInput,      // input buffer handle 
                irInBuf,     // buffer to read into 
                128,         // size of read buffer 
                &cNumRead) ) // number of records read 
            break; 
 
        // Dispatch the events to the appropriate handler. 
 
        for (i = 0; i < cNumRead; i++) 
        {
            switch(irInBuf[i].EventType) 
            { 
                case KEY_EVENT: // keyboard input 
					
                    DBGConsole_ProcessKeyEvent(irInBuf[i].Event.KeyEvent); 

                    break; 
 
               /* case MOUSE_EVENT: // mouse input 
                    MouseEventProc(irInBuf[i].Event.MouseEvent); 
                    break; */
 
                /*case WINDOW_BUFFER_SIZE_EVENT: // scrn buf. resizing 
                    ResizeEventProc( 
                        irInBuf[i].Event.WindowBufferSizeEvent); 
                    break; */
 
                case FOCUS_EVENT:  // disregard focus events 
 
                case MENU_EVENT:   // disregard menu events 
                    break; 
 
                default: 
                    break; 
            } 
		}

	}

	return 0;
}
	


BOOL DBGConsole_ResizeConBufAndWindow(HANDLE hConsole, SHORT xSize, SHORT ySize)
{
	CONSOLE_SCREEN_BUFFER_INFO csbi; /* hold current console buffer info */
	BOOL bSuccess;
	SMALL_RECT srWindowRect; /* hold the new console size */
	COORD coordScreen;

	bSuccess = GetConsoleScreenBufferInfo(hConsole, &csbi);
	if (!bSuccess)
		return FALSE;

	/* get the largest size we can size the console window to */

	coordScreen = GetLargestConsoleWindowSize(hConsole); 
	if (!bSuccess)
		return FALSE;

	/* define the new console window size and scroll position */
	srWindowRect.Right = (SHORT) (min(xSize, coordScreen.X) - 1);
	srWindowRect.Bottom = (SHORT) (min(ySize, coordScreen.Y) - 1);

	srWindowRect.Left = srWindowRect.Top = (SHORT) 0;

	/* define the new console buffer size */
	coordScreen.X = xSize;
	coordScreen.Y = ySize;

	/* if the current buffer is larger than what we want, resize the */
	/* console window first, then the buffer */
	if ((DWORD) csbi.dwSize.X * csbi.dwSize.Y > (DWORD) xSize * ySize)
	{
		bSuccess = SetConsoleWindowInfo(hConsole, TRUE, &srWindowRect);
		if (!bSuccess)
			return FALSE;

		bSuccess = SetConsoleScreenBufferSize(hConsole, coordScreen);
		if (!bSuccess)
			return FALSE;
	}

	/* if the current buffer is smaller than what we want, resize the */
	/* buffer first, then the console window */
	if ((DWORD) csbi.dwSize.X * csbi.dwSize.Y < (DWORD) xSize * ySize) 
	{
		bSuccess = SetConsoleScreenBufferSize(hConsole, coordScreen);
		if (!bSuccess)
			return FALSE;

		bSuccess = SetConsoleWindowInfo(hConsole, TRUE, &srWindowRect);
		if (!bSuccess)
			return FALSE;
	}

	/* if the current buffer *is* the size we want, don't do anything! */

	return TRUE;
} 

void DBGConsole_DumpRegisters()
{
	if (s_nMidPaneType == DBGCON_RSP_CODE_VIEW)
		DBGConsole_DumpRSPRegisters();
	else
		DBGConsole_DumpCPURegisters();

}

void DBGConsole_ClearRegisters()
{
	BOOL bSuccess;
    COORD coord; 
	DWORD cWritten;

	for (coord.Y = 0; coord.Y < 21; coord.Y++)
	{
		coord.X = 0;            // start at first cell 

		// Write a string of colors to a screen buffer. 
		bSuccess = FillConsoleOutputAttribute(
			g_hStdOut,
			sc_wAttrWhite,
			80,		// Clear to end
			coord,
			&cWritten);

		bSuccess = FillConsoleOutputCharacter(g_hStdOut,
			' ', 80, coord, &cWritten);


	}


}

void DBGConsole_DumpCPURegisters()
{
	char msg[200];

	DWORD base;
	DWORD dwReg;
	SHORT wX;
	SHORT wY;
	static BOOL bInit = FALSE;
	static u64 qwGPRCopy[32];

	static const DWORD dwRegDisplayWidth = 11;

	// Make sure that we render all registers - set all regs to 
	if (bInit == FALSE)
	{
		for (dwReg = 0; dwReg < 32; dwReg++)
		{
			qwGPRCopy[dwReg] = g_qwGPR[dwReg];
		}
		bInit = TRUE;
	}

	wX = 0;
	wY = sc_nTopPaneTextOffset+1;

	sprintf(msg, "PC:%08x", g_dwPC);
	DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, 19); wY ++;
	sprintf(msg, "MH:%016I64x", g_qwMultHI);
	DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, 19); wY ++;
	sprintf(msg, "ML:%016I64x", g_qwMultLO);
	DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, 19); wY ++;

	// Do main registers
	for (base = 0; base < 32; base += 8)
	{
		wY = sc_nTopPaneTextOffset+1+3;
		for (dwReg = base; dwReg < base+8; dwReg++)
		{
			// Not wsprintf as we want long int output
			wsprintf(msg, "%s:%08x", RegNames[dwReg], (DWORD)g_qwGPR[dwReg]);
			
			// Only draw if it's changed
			if (qwGPRCopy[dwReg] != g_qwGPR[dwReg])
			{
				DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrBoldRed, dwRegDisplayWidth);
				qwGPRCopy[dwReg] = g_qwGPR[dwReg];
			}
			else
			{
				DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, dwRegDisplayWidth);
			}
			wY++;
		}
		wX += dwRegDisplayWidth + 1;
	}
	
	// Do CoPro1 regs
	wY = sc_nTopPaneTextOffset+1;
	DWORD i;
	for (i = 0; ; i++)
	{
		dwReg = sc_Cop0RegsToShow[i];
		if (dwReg == ~0)
			break;

		sprintf(msg, "%4.4s:%08x", ShortCop0RegNames[dwReg], (DWORD)g_qwCPR[0][dwReg]);
		DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, 13);
		wY ++;
		if (i == 10)
		{
			wX += 15;
			wY = sc_nTopPaneTextOffset+1;
		}
	}


}


void DBGConsole_DumpRSPRegisters()
{
	char msg[200];

	DWORD dwReg;
	DWORD base;
	WORD wX = 0;
	WORD wY = 0;
	static BOOL bInit = FALSE;
	static DWORD dwRSPGPRCopy[32];

	static const DWORD dwRegDisplayWidth = 15;


	if (bInit == FALSE)
	{
		// Make sure that we render all registers
		for (dwReg = 0; dwReg < 32; dwReg++) {
			dwRSPGPRCopy[dwReg] = g_dwRSPGPR[dwReg];
		}
		bInit = TRUE;
	}

	wY = sc_nTopPaneTextOffset+1;
	wsprintf(msg, "PC: 0x%08x", RSPPC());
	DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, dwRegDisplayWidth+8); wY++;

	// Do main registers
	for (base = 0; base < 32; base += 8)
	{
		wY = sc_nTopPaneTextOffset+1+3;
		for (dwReg = base; dwReg < base+8; dwReg++) {
			wsprintf(msg, "%s: 0x%08x", RegNames[dwReg], g_dwRSPGPR[dwReg]);

			// Only draw if changed
			if (dwRSPGPRCopy[dwReg] != g_dwRSPGPR[dwReg])
			{
				DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrBoldRed, 20);
				dwRSPGPRCopy[dwReg] = g_dwRSPGPR[dwReg];
			}
			else
			{
				DBGConsole_WriteString(msg, FALSE, wX, wY, sc_wAttrWhite, 20);
			}
			wY++;
		}
		wX += dwRegDisplayWidth + 1;
	}
	
}


void DBGConsole_SetActivePane(DBGCON_PANETYPE pane)
{
	
	s_nActivePane = pane;

	DBGConsole_RedrawPaneLabels();

}

void DBGConsole_RedrawPaneLabels()
{
	int i;
	CHAR szLine[80+1];
	CHAR szPaneName[80+1];
	CHAR * s, * d;
	static const CHAR cLine = (char)196;
	static const CHAR cDownArrow = 31;		

	// Top pane
	lstrcpyn(szPaneName, "Registers", 80);
	for (i = 0; i < 80; i++)
		szLine[i] = cLine;

	if (s_nActivePane == DBGCON_REG_PANE)
	{
		szLine[1] = szLine[2] = szLine[3] = szLine[4] = cDownArrow;
	}
	szLine[5] = 'F'; szLine[6] = '1';
	for (s = szPaneName, d = &szLine[8]; s[0] != 0; s++,d++)
		d[0] = s[0];
	szLine[80] = 0;
	DBGConsole_WriteString(szLine, FALSE, 0, sc_nTopPaneTextOffset, BACKGROUND_GREEN, 80);

	// Mem pane
	lstrcpyn(szPaneName, "Memory", 80);
	for (i = 0; i < 80; i++)
		szLine[i] = cLine;

	if (s_nActivePane == DBGCON_MEM_PANE)
	{
		szLine[1] = szLine[2] = szLine[3] = szLine[4] = cDownArrow;
	}
	szLine[5] = 'F'; szLine[6] = '2';
	for (s = szPaneName, d = &szLine[8]; s[0] != 0; s++,d++)
		d[0] = s[0];
	szLine[80] = 0;
	DBGConsole_WriteString(szLine, FALSE, 0, sc_nMemPaneTextOffset, BACKGROUND_GREEN, 80);


	// Mid pane
	wsprintf(szPaneName, "%s", s_szPaneNameLabels[s_nMidPaneType]);
	for (i = 0; i < 80; i++)
		szLine[i] = cLine;

	if (s_nActivePane == DBGCON_MID_PANE)
	{
		szLine[1] = szLine[2] = szLine[3] = szLine[4] = cDownArrow;
	}
	szLine[5] = 'F'; szLine[6] = '3';
	for (s = szPaneName, d = &szLine[8]; s[0] != 0; s++,d++)
		d[0] = s[0];
	szLine[80] = 0;
	DBGConsole_WriteString(szLine, FALSE, 0, sc_nMidPaneTextOffset, BACKGROUND_GREEN, 80);


	// Bottom pane:
	lstrcpyn(szPaneName, "Output", 80);
	for (i = 0; i < 80; i++)
		szLine[i] = cLine;

	if (s_nActivePane == DBGCON_BOT_PANE)
	{
		szLine[1] = szLine[2] = szLine[3] = szLine[4] = cDownArrow;
	}
	szLine[5] = 'F'; szLine[6] = '4';
	for (s = szPaneName, d = &szLine[8]; s[0] != 0; s++,d++)
		d[0] = s[0];
	szLine[80] = 0;
	DBGConsole_WriteString(szLine, FALSE, 0, sc_nBotPaneTextOffset, BACKGROUND_GREEN, 80);

}

void DBGConsole_DumpContext(DWORD dwPC)
{
	switch (s_nMidPaneType)
	{
	case DBGCON_CPU_CODE_VIEW:
		DBGConsole_DumpCPUContext(dwPC);
		break;
	case DBGCON_RSP_CODE_VIEW:
		DBGConsole_DumpRSPContext(dwPC);
		break;
	}

}

void DBGConsole_DumpContext_PrevLine()
{
	switch (s_nActivePane)
	{
	case DBGCON_MEM_PANE:
		DBGConsole_DumpMemory(g_dwMemAddress - 4*4);
		break;
	case DBGCON_MID_PANE:
		switch (s_nMidPaneType)
		{
		case DBGCON_CPU_CODE_VIEW:
			DBGConsole_DumpCPUContext(g_dwDisplayPC - 4);
			break;
		case DBGCON_RSP_CODE_VIEW:
			DBGConsole_DumpRSPContext(g_dwDisplayRSP_PC - 4);
			break;

		}
		break;
	case DBGCON_BOT_PANE:
		DBGConsole_DisplayOutput(g_nOutputBufOffset - 1);
		break;
	}
}

void DBGConsole_DumpContext_PrevPage()
{
	switch (s_nActivePane)
	{
	case DBGCON_MEM_PANE:
		DBGConsole_DumpMemory(g_dwMemAddress - 4*4*sc_nMidPaneLines);
		break;
	case DBGCON_MID_PANE:
		switch (s_nMidPaneType)
		{
		case DBGCON_CPU_CODE_VIEW:
			DBGConsole_DumpCPUContext(g_dwDisplayPC - 4*sc_nMidPaneLines);
			break;
		case DBGCON_RSP_CODE_VIEW:
			DBGConsole_DumpRSPContext(g_dwDisplayRSP_PC - 4*sc_nMidPaneLines);
			break;
		}
		break;
	case DBGCON_BOT_PANE:
		DBGConsole_DisplayOutput(g_nOutputBufOffset - sc_nBotPaneLines);
		break;
	}
}

void DBGConsole_DumpContext_NextLine()
{
	switch (s_nActivePane)
	{
	case DBGCON_MEM_PANE:
		DBGConsole_DumpMemory(g_dwMemAddress + 4*4);
		break;
	case DBGCON_MID_PANE:
		switch (s_nMidPaneType)
		{
		case DBGCON_CPU_CODE_VIEW:
			DBGConsole_DumpCPUContext(g_dwDisplayPC + 4);
			break;
		case DBGCON_RSP_CODE_VIEW:
			DBGConsole_DumpRSPContext(g_dwDisplayRSP_PC + 4);
			break;
		}
		break;
	case DBGCON_BOT_PANE:
		DBGConsole_DisplayOutput(g_nOutputBufOffset + 1);
		break;
	}

}

void DBGConsole_DumpContext_NextPage()
{
	switch (s_nActivePane)
	{
	case DBGCON_MEM_PANE:
		DBGConsole_DumpMemory(g_dwMemAddress + 4*4*sc_nMidPaneLines);
		break;
	case DBGCON_MID_PANE:
		switch (s_nMidPaneType)
		{
		case DBGCON_CPU_CODE_VIEW:
			DBGConsole_DumpCPUContext(g_dwDisplayPC + 4*sc_nMidPaneLines);
			break;
		case DBGCON_RSP_CODE_VIEW:
			DBGConsole_DumpRSPContext(g_dwDisplayRSP_PC + 4*sc_nMidPaneLines);
			break;
		}
		break;
	case DBGCON_BOT_PANE:
		DBGConsole_DisplayOutput(g_nOutputBufOffset + sc_nBotPaneLines);
		break;
	}
}

// Pass in ~0 to display current mem pointer
void DBGConsole_DumpMemory(DWORD dwAddress)
{
	WORD wY;
	char buf[120];

	if (dwAddress == ~0)
		dwAddress = g_dwMemAddress;

	g_dwMemAddress = dwAddress;

	// Clear the background area...

	wY = sc_nMemPaneTextOffset+1;
	for (DWORD dwIndex = 0; dwIndex < sc_nMemPaneLines; dwIndex++)
	{
		DWORD * pMem;
		DWORD dwBase = InternalReadAddress(dwAddress, (LPVOID*)&pMem);

		// TODO - Lookup a base for each of pMem[0], pMem[1] etc - this
		// is significant if we cross a boundary, but not *really* important

		if (dwBase != MEM_UNUSED)
		{
			wsprintf(buf, "0x%08x: %08x %08x %08x %08x ", dwAddress, pMem[0], pMem[1], pMem[2], pMem[3]);					

			char * p = buf + lstrlen(buf);
			// Concat characters
			for (int i = 0; i < 16; i++)
			{
				char c = ((BYTE*)pMem)[i ^ 0x3];
				if (isprint(c))
					*p++ = c;
				else 
					*p++ = '.';

				if ((i%4)==3)
					*p++ = ' ';
			}
			*p++ = 0;
		}
		else
		{
			wsprintf(buf, "0x%08x: ???????? ???????? ???????? ???????? .... .... .... ....", dwAddress);
		}

		DBGConsole_WriteString(buf, FALSE, 0, wY, sc_wAttrWhite, 80);
		wY++;

		dwAddress+=4*4;
	}
}



// Pass in ~0 to display current PC
void DBGConsole_DumpCPUContext(DWORD dwPC)
{
	WORD wY;
	char opinfo[120];
	char msg[200];

	if (dwPC == ~0)
		dwPC = g_dwPC;

	g_dwDisplayPC = dwPC;

	dwPC -= (4*(sc_nMidPaneLines/2));	// Show previous instructions

	// Clear the background area...

	wY = sc_nMidPaneTextOffset+1;
	for (DWORD dwIndex = 0; dwIndex < sc_nMidPaneLines; dwIndex++)
	{
		DWORD * pPC;
		DWORD dwBase = InternalReadAddress(dwPC, (LPVOID*)&pPC);
		WORD wAttr;

		if (dwPC == g_dwPC) 
			wAttr = sc_wAttrBoldWhite;
		else
			wAttr = sc_wAttrWhite;

		if (dwBase != MEM_UNUSED)
		{
			DWORD dwOpCode = pPC[0];
			LPCSTR pszSymbol;

			SprintOpCodeInfo(opinfo, dwPC, dwOpCode);
			if (op(dwOpCode) == OP_DBG_BKPT)
			{
				wsprintf(msg, "[R!] 0x%08x: %s", dwPC, opinfo);
			}
			else
			{
				wsprintf(msg, "  0x%08x: %s", dwPC, opinfo);
			}
						
			//if (op(dwOpCode) == OP_PATCH)
			{
				pszSymbol = Patch_GetJumpAddressName(dwPC);
				if (pszSymbol[0] != '?')
				{
					// This is actually a patch target
					DBGConsole_WriteString("", FALSE, 0, wY, wAttr, 80);
					wY++;
					dwIndex++;

					if (dwIndex < sc_nMidPaneLines)
					{
						TCHAR szFuncInfo[30];
						wsprintf(szFuncInfo, "[G%s]", pszSymbol);
						DBGConsole_WriteString(szFuncInfo, TRUE, 0, wY, wAttr, 80);
						wY++;
						dwIndex++;
					}
				}
			}
		}
		else
		{
			wsprintf(msg, "0x%08x: ???", dwPC);
		}


		if (dwIndex < sc_nMidPaneLines)
		{
			DBGConsole_WriteString(msg, TRUE, 0, wY, wAttr, 80);
			wY++;
		}

		dwPC+=4;
	}
}

void DBGConsole_DumpRSPContext(DWORD dwPC)
{
	char opinfo[256];
	char msg[200];
	WORD wY;

	DWORD dwRSPPC = RSPPC() | 0x04001000;

	if (dwPC == ~0)
		dwPC = dwRSPPC;

	g_dwDisplayRSP_PC = dwPC;

	dwPC -= (4*(sc_nMidPaneLines/2));	// Show previous 4 instructions

	// Clear the background area...

	wY = sc_nMidPaneTextOffset+1;
	for (DWORD dwIndex = 0; dwIndex < sc_nMidPaneLines; dwIndex++)
	{
		if (dwPC >= 0x04002000)
			break;

		//- TODO - Translate to InternalRead
		DWORD dwOpCode = Read32Bits(PHYS_TO_K1(dwPC));

		SprintRSPOpCodeInfo(opinfo, dwPC, dwOpCode);
		wsprintf(msg, "0x%08x: <0x%08x> %s", dwPC, dwOpCode, opinfo);

		if ((dwPC&0x1FFF) == (dwRSPPC&0x1FFF))
		{
			DBGConsole_WriteString(msg, FALSE, 0, wY, sc_wAttrBoldWhite, 80);
		} else {
			DBGConsole_WriteString(msg, FALSE, 0, wY, sc_wAttrWhite, 80);
		}
		wY++;
		dwPC += 4;
		
	}
}



// Write a string of characters to a screen buffer. 
BOOL DBGConsole_WriteString(LPCTSTR szString, BOOL bParse, SHORT x, SHORT y, WORD wAttr, SHORT width)
{
    DWORD cWritten; 
    BOOL bSuccess; 
    COORD coord; 
	WORD wNumToWrite;
	TCHAR szBuffer[2048+1];
	WORD arrwAttributes[2048];

    coord.X = x;            // start at first cell 
    coord.Y = y;            //   of first row      

	lstrcpyn(szBuffer, szString, 2048);

	if (bParse)
		DBGConsole_ParseStringHighlights(szBuffer, arrwAttributes, wAttr);

	wNumToWrite = lstrlen(szBuffer);
	if (wNumToWrite > width)
		wNumToWrite = width;

    bSuccess = WriteConsoleOutputCharacter( 
        g_hStdOut,              // screen buffer handle 
        szBuffer,				// pointer to source string 
        wNumToWrite,			// length of string 
        coord,					// first cell to write to 
        &cWritten);				// actual number written 
    if (!bSuccess) 
        return FALSE; 
 
	// Copy the visible portion to the display
	if (bParse)
	{
		/*bSuccess = */WriteConsoleOutputAttribute(
			g_hStdOut,
			arrwAttributes,
			wNumToWrite,
			coord,
			&cWritten);
	}
	else
	{
		// Just clear
		/*bSuccess = */FillConsoleOutputAttribute(
			g_hStdOut,
			wAttr,
			wNumToWrite,		// Clear to end
			coord,
			&cWritten);
	}

	// Clear the end of the buffer
	if (coord.X + wNumToWrite < width)
	{
		coord.X += wNumToWrite;
		/*bSuccess = */FillConsoleOutputAttribute(
			g_hStdOut,
			wAttr,
			width - wNumToWrite,		// Clear to end
			coord,
			&cWritten);

		/*bSuccess = */FillConsoleOutputCharacter(g_hStdOut,
			' ', width - coord.X, coord, &cWritten);
	}
	
	return TRUE;
}





void DBGConsole_ProcessKeyEvent(KEY_EVENT_RECORD kr)
{
	TCHAR szInputChar[1+1];
	COORD curPos;
	DWORD dwInputLen;

	dwInputLen = lstrlen(g_szInput);

	if (kr.bKeyDown)
	{
		switch (kr.wVirtualKeyCode)
		{
		case VK_RETURN:
			// Process the current input
			if (lstrlen(g_szInput) > 0)
			{
				DBGConsole_ProcessInput();

				// Reset the input buffer
				lstrcpy(g_szInput, "");

				DBGConsole_WriteString(g_szInput, FALSE, 0, sc_nCommandTextOffset, BACKGROUND_GREEN | BACKGROUND_BLUE, 80);
			}
			break;
		case VK_BACK:
			if (dwInputLen > 0)
			{
				g_szInput[dwInputLen-1] = '\0';
			}
			DBGConsole_WriteString(g_szInput, FALSE, 0, sc_nCommandTextOffset, BACKGROUND_GREEN | BACKGROUND_BLUE, 80);
			break;
		case VK_PRIOR:
			DBGConsole_DumpContext_PrevPage();
			break;
		case VK_UP:
			DBGConsole_DumpContext_PrevLine();
			break;
		case VK_HOME:
		/*	if (s_bShowCPU)
				DBGConsole_DumpCPUContext(~0);
			else
				DBGConsole_DumpRSPContext(~0);*/
			break;
		case VK_DOWN:
			DBGConsole_DumpContext_NextLine();
			break;
		case VK_NEXT:
			DBGConsole_DumpContext_NextPage();
			break;

		case VK_F1:
			DBGConsole_SetActivePane(DBGCON_REG_PANE);
			break;
		case VK_F2:
			DBGConsole_SetActivePane(DBGCON_MEM_PANE);
			break;
		case VK_F3:
			DBGConsole_SetActivePane(DBGCON_MID_PANE);
			break;
		case VK_F4:
			DBGConsole_SetActivePane(DBGCON_BOT_PANE);
			break;

		// Step into
		case VK_F10:
			CPUStep();
			DBGConsole_UpdateDisplay();
			break;

		default:
			if (kr.wVirtualKeyCode == VK_SPACE ||
				kr.uChar.AsciiChar == '_' || 
				kr.uChar.AsciiChar == '[' || 
				kr.uChar.AsciiChar == ']' || 
				
				kr.uChar.AsciiChar == '+' || 
				kr.uChar.AsciiChar == '-' || 
								
				(kr.wVirtualKeyCode >= '0' &&
				 kr.wVirtualKeyCode <= '9') ||

				(kr.wVirtualKeyCode >= 'A' &&
				kr.wVirtualKeyCode <= 'Z'))
			{
			
				wsprintf(szInputChar, "%c", kr.uChar.AsciiChar);
				lstrcat(g_szInput, szInputChar);

				DBGConsole_WriteString(g_szInput, FALSE, 0, sc_nCommandTextOffset, BACKGROUND_GREEN | BACKGROUND_BLUE, 80);
			}

			break;
		}
	}

	curPos.X = lstrlen(g_szInput);
	curPos.Y = sc_nCommandTextOffset;
						
	SetConsoleCursorPosition(g_hStdOut, curPos);

}



void DBGConsole_ProcessInput()
{
	DWORD i;
	BOOL bProcessed;

	bProcessed = FALSE;

	for (i = 0; ; i++)
	{
		if (g_DBGCommands[i].szCommand == NULL)
			break;

		DWORD dwCmdLen = lstrlen(g_DBGCommands[i].szCommand);
		DWORD dwInputLen = lstrlen(g_szInput);

		if (strnicmp(g_DBGCommands[i].szCommand, g_szInput, dwCmdLen) == 0)
		{
			if (g_DBGCommands[i].pArglessFunc != NULL)
				g_DBGCommands[i].pArglessFunc();
			else if (g_DBGCommands[i].pArgFunc != NULL)
			{
				// Trim spaces...
				LPTSTR szArgs = g_szInput + dwCmdLen;
				while (*szArgs == ' ')
					szArgs++;

				g_DBGCommands[i].pArgFunc(szArgs);
			}
			bProcessed = TRUE;
			break;
		}
	}

	if (bProcessed == FALSE)
		DBGConsole_Msg(0, "[YUnknown command: %s]", g_szInput);
}

void DBGCommand_Mem(LPCTSTR szCommand)
{
	DWORD dwAddress;
	DWORD dwOffset;
	LONG n;

	dwOffset = 0;
	
	// Check if this is a register, e.g. "mem [sp+30]
	if (szCommand[0] == '[')
	{
		TCHAR szRegName[100];
		TCHAR * pOffset;
		int i;
		int reg;

		for (i = 0; szCommand[1+i] != 0 &&
					szCommand[1+i] != ']'; i++)
		{
			szRegName[i] = szCommand[1+i];
		}
		szRegName[i] = 0;

		pOffset = strchr(szRegName, '+');
		if (pOffset == NULL)
			pOffset = strchr(szRegName, '-');
		if (pOffset != NULL)
		{
			// This is of the form [sp+12]
			n = sscanf(pOffset, "%x", &dwOffset);
			if (n == 0)
				dwOffset = 0;

			// Bung in null-termiate
			pOffset[0] = 0;
			// Now two strings, "sp" and "12"
		}

		// find register
		reg = -1;
		for (i = 0; i < 32; i++)
		{
			if (lstrcmpi(szRegName, RegNames[i]) == 0)
			{
				reg = i;
				break;
			}
		}
		if (reg != -1)
		{
			dwAddress = (DWORD)g_qwGPR[reg];
			n = 1;
		}
		else
		{
			DBGConsole_Msg(0, "[Yreg %s not found]", szRegName);
			n = 0;
		}
	}
	else
	{
		n = sscanf(szCommand, "0x%x", &dwAddress);
		if (n != 1)
			// Check for absence of 0x prefix - always hex!
			n = sscanf(szCommand, "%x", &dwAddress);
	}


	if (n != 1)
	{
		DBGConsole_Msg(0, "Invalid argument");
	}
	else
	{
		DBGConsole_DumpMemory(dwAddress + dwOffset);
	}

}
void DBGCommand_List(LPCTSTR szCommand)
{
	DWORD dwAddress;
	LONG n = sscanf(szCommand, "0x%08x", &dwAddress);

	if (n != 1)
	{
		DBGConsole_Msg(0, "Invalid argument");
	}
	else
	{
		s_nMidPaneType = DBGCON_CPU_CODE_VIEW;
		DBGConsole_RedrawPaneLabels();
		DBGConsole_DumpCPUContext(dwAddress);
	}

}


void DBGCommand_ListOS(LPCTSTR szCommand)
{
	if (lstrlen(szCommand) == 0)
	{
		DBGConsole_Msg(0, "Invalid argument: %s", szCommand);
	}
	else
	{
		s_nMidPaneType = DBGCON_CPU_CODE_VIEW;
		DBGConsole_RedrawPaneLabels();

		DWORD dwAddress = Patch_GetSymbolAddress(szCommand);
		if (dwAddress != ~0)
			DBGConsole_DumpCPUContext(dwAddress);
		else
			DBGConsole_Msg(0, "Symbol %s not found", szCommand);
	}

}

void DBGCommand_BPX(LPCTSTR szCommand)
{
	DWORD dwAddress;
	LONG n = sscanf(szCommand, "0x%08x", &dwAddress);

	if (n != 1)
	{
		DBGConsole_Msg(0, "Invalid argument: %s", szCommand);
	}
	else
	{		
		CPU_AddBreakPoint(dwAddress);
		DBGConsole_UpdateDisplay();
	}
}

void DBGCommand_ShowCPU()
{
	s_nMidPaneType = DBGCON_CPU_CODE_VIEW;
	DBGConsole_ClearRegisters();
	DBGConsole_DumpContext(~0);
	DBGConsole_DumpRegisters();
	DBGConsole_RedrawPaneLabels();
}
void DBGCommand_ShowRSP()
{
	s_nMidPaneType = DBGCON_RSP_CODE_VIEW;
	DBGConsole_ClearRegisters();
	DBGConsole_DumpContext(~0);
	DBGConsole_DumpRegisters();
	DBGConsole_RedrawPaneLabels();
}


void DBGConsole_HelpCommand()
{
	DBGConsole_Msg(0, "Help will be available soon!");

	DBGConsole_UpdateDisplay();
}


void DBGConsole_GoCommand()
{
	if (StartCPUThread() == FALSE)
	{
		DBGConsole_Msg(0, "Unable to start CPU thread, or thread already running");
	}
}

void DBGConsole_StopCommand()
{
	StopCPUThread();
}

void DBGConsole_ParseStringHighlights(LPSTR szString, WORD * arrwAttributes, WORD wAttr)
{
	WORD wCurrAttribute = wAttr;
	int iIn, iOut;
	int nMax = lstrlen(szString);

	for (iIn = 0, iOut = 0; iIn < nMax; iIn++)
	{
		if (szString[iIn] == '[')
		{
			switch (szString[iIn+1])
			{
			case 'r': wCurrAttribute = FOREGROUND_RED; break;
			case 'g': wCurrAttribute = FOREGROUND_GREEN; break;
			case 'b': wCurrAttribute = FOREGROUND_BLUE; break;
			case 'c': wCurrAttribute = FOREGROUND_GREEN|FOREGROUND_BLUE; break;
			case 'm': wCurrAttribute = FOREGROUND_RED|FOREGROUND_BLUE; break;
			case 'y': wCurrAttribute = FOREGROUND_RED|FOREGROUND_GREEN; break;
			case 'w': wCurrAttribute = sc_wAttrWhite; break;
			case 'R': wCurrAttribute = FOREGROUND_RED|FOREGROUND_INTENSITY; break;
			case 'G': wCurrAttribute = FOREGROUND_GREEN|FOREGROUND_INTENSITY; break;
			case 'B': wCurrAttribute = FOREGROUND_BLUE|FOREGROUND_INTENSITY; break;
			case 'C': wCurrAttribute = FOREGROUND_GREEN|FOREGROUND_BLUE|FOREGROUND_INTENSITY; break;
			case 'M': wCurrAttribute = FOREGROUND_RED|FOREGROUND_BLUE|FOREGROUND_INTENSITY; break;
			case 'Y': wCurrAttribute = FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY; break;
			case 'W': wCurrAttribute = sc_wAttrBoldWhite; break;
			}

			// Skip colour character
			iIn++;
		}
		else if (szString[iIn] == ']')
		{
			wCurrAttribute = wAttr;
		}
		else
		{
			szString[iOut] = szString[iIn];
			arrwAttributes[iOut] = wCurrAttribute;

			iOut++;
		}
	}
	szString[iOut] = '\0';

}

void __cdecl DBGConsole_Stats(DWORD dwRow, LPCTSTR szFormat, ...)
{

    DWORD cWritten; 
    BOOL bSuccess; 
    COORD coord; 
	WORD wNumToWrite;
    va_list va;
	TCHAR szBuffer[2048+1];

	// Return if we don't have a console!
	if (!g_bShowDebug || !s_bConsoleAllocated)
		return;


    coord.X = 50;
    coord.Y = (SHORT)(25+dwRow);
	
	WORD wMaxWidth = 80 - coord.X;

	// Format the output
	va_start(va, szFormat);
	// Don't use wvsprintf as it doesn't handle floats!
	vsprintf(szBuffer, szFormat, va);
	va_end(va);


	wNumToWrite = lstrlen(szBuffer);
	if (wNumToWrite > wMaxWidth)
		wNumToWrite = wMaxWidth;

    bSuccess = WriteConsoleOutputCharacter( 
        g_hStdOut,              // screen buffer handle 
        szBuffer,								// pointer to source string 
        wNumToWrite,							// length of string 
        coord,									// first cell to write to 
        &cWritten);								// actual number written 
    if (!bSuccess) 
        return; 
 
	// Write a string of colors to a screen buffer. 
	bSuccess = FillConsoleOutputAttribute(
		g_hStdOut,
		sc_wAttrWhite,
		wMaxWidth,		// Clear to end
		coord,
		&cWritten);
    if (!bSuccess) 
        return;  

	// Clear the end of the buffer
	if (wNumToWrite < wMaxWidth)
	{
		coord.X += wNumToWrite;
		FillConsoleOutputCharacter(g_hStdOut,
			' ', wMaxWidth - wNumToWrite, coord, &cWritten);
	}
	
	return;


}



void __cdecl DBGConsole_Msg(DWORD dwType, LPCTSTR szFormat, ...)
{
	BOOL bSuccess; 
	CONSOLE_SCREEN_BUFFER_INFO csbiInfo; 
	SMALL_RECT srctScrollRect, srctClipRect; 
	CHAR_INFO chiFill; 
	COORD coordDest; 
	DWORD cWritten;
    va_list va;
	TCHAR szBuffer[2048+1];
	LONG nOutputLength;

	WORD arrwAttributes[2048];

	// Return if we don't have a console!
	if (!g_bShowDebug || !s_bConsoleAllocated)
		return;

	if (szFormat == NULL)
	{
		MessageBox(g_hMainWindow, "Null string passed to DBGConsole_Msg!", g_szDaedalusName, MB_OK);
		return;
	}
    
	bSuccess = GetConsoleScreenBufferInfo(g_hOutputBuffer, &csbiInfo); 
	if (!bSuccess) 
		return;

	// Parse the buffer:
	try {
		// Format the output
		va_start(va, szFormat);
		// Don't use wvsprintf as it doesn't handle floats!
		vsprintf(szBuffer, szFormat, va);
		va_end(va);
		// Scan through szBuffer and set up attributes
		DBGConsole_ParseStringHighlights(szBuffer, arrwAttributes, sc_wAttrWhite);

		nOutputLength = lstrlen(szBuffer);
	}
	catch (...)
	{
		return;
	}
	
	// Scroll the contents of the buffer up by one line
	// Define the scrolling rectangle (we lose the top line)
	srctScrollRect.Top = 1; 
	srctScrollRect.Bottom = csbiInfo.dwSize.Y - 1; 
	srctScrollRect.Left = 0; 
	srctScrollRect.Right = csbiInfo.dwSize.X - 1; 
	
	// The destination for the scroll rectangle is one row up. 
	coordDest.X = 0; 
	coordDest.Y = 0; 
	
	// The clipping rectangle is the same as the scrolling rectangle. 
	// The destination row is left unchanged.
	srctClipRect = srctScrollRect; 
	
	// Fill the bottom row
	chiFill.Attributes = sc_wAttrWhite; 
	chiFill.Char.AsciiChar = ' '; 
	
	// Scroll up one line. 
	/*bSuccess = */ScrollConsoleScreenBuffer( 
		g_hOutputBuffer,         // screen buffer handle 
		&srctScrollRect, // scrolling rectangle 
		&srctClipRect,   // clipping rectangle 
		coordDest,       // top left destination cell 
		&chiFill);       // fill character and color 


	// Write the output to the buffer
	COORD coordCursor;

	// Write the new string to the bottom of the buffer
	coordCursor.X = 0;
	coordCursor.Y = csbiInfo.dwSize.Y - 1; 

	/*bSuccess = */WriteConsoleOutputCharacter( 
		g_hOutputBuffer,            // screen buffer handle 
		szBuffer,					// pointer to source string 
		nOutputLength,				// length of string 
		coordCursor,				// first cell to write to 
		&cWritten);					// actual number written 

	/*bSuccess = */WriteConsoleOutputAttribute(
		g_hOutputBuffer,
		arrwAttributes,
		nOutputLength,
		coordCursor,
		&cWritten);

	DBGConsole_DisplayOutput(0);

}

// Used to overwrite previous lines - like for doing % complete indicators
void __cdecl DBGConsole_MsgOverwrite(DWORD dwType, LPCTSTR szFormat, ...)
{
	BOOL bSuccess; 
	CONSOLE_SCREEN_BUFFER_INFO csbiInfo; 
	DWORD cWritten;
    va_list va;
	TCHAR szBuffer[2048+1];
	LONG nOutputLength;

	WORD arrwAttributes[2048];

	// Return if we don't have a console!
	if (!g_bShowDebug || !s_bConsoleAllocated)
		return;

    
	bSuccess = GetConsoleScreenBufferInfo(g_hOutputBuffer, &csbiInfo); 
	if (!bSuccess) 
		return;

	// Parse the buffer:
	{
		// Format the output
		va_start(va, szFormat);
		// Don't use wvsprintf as it doesn't handle floats!
		vsprintf(szBuffer, szFormat, va);
		va_end(va);
		// Scan through szBuffer and set up attributes
		DBGConsole_ParseStringHighlights(szBuffer, arrwAttributes, sc_wAttrWhite);

		nOutputLength = lstrlen(szBuffer);

	}
	
	// Write the output to the buffer
	{
		COORD coord;

		// Write the new string to the bottom of the buffer
		coord.X = 0;
		coord.Y = csbiInfo.dwSize.Y - 1; 

		/*bSuccess = */WriteConsoleOutputCharacter( 
			g_hOutputBuffer,            // screen buffer handle 
			szBuffer,					// pointer to source string 
			nOutputLength,				// length of string 
			coord,						// first cell to write to 
			&cWritten);					// actual number written 

		/*bSuccess = */WriteConsoleOutputAttribute(
			g_hOutputBuffer,
			arrwAttributes,
			nOutputLength,
			coord,
			&cWritten);

			// Clear the end of the buffer
			if (nOutputLength < 80)
			{
				coord.X += nOutputLength;
				FillConsoleOutputCharacter(g_hOutputBuffer,
					' ', 80 - nOutputLength, coord, &cWritten);
			}
	}

	DBGConsole_DisplayOutput(0);


}



// Copy the visible portion to the display
void DBGConsole_DisplayOutput(LONG nOffset)
{
	CONSOLE_SCREEN_BUFFER_INFO csbiInfo; 
	SMALL_RECT srctReadRect; 
	SMALL_RECT srctWriteRect; 
	CHAR_INFO chiBuffer[sc_nBotPaneLines*80];
	COORD coordBufSize; 
	COORD coordBufCoord; 
	BOOL bSuccess; 

	bSuccess = GetConsoleScreenBufferInfo(g_hOutputBuffer, &csbiInfo); 
	if (!bSuccess) 
		return;


	// We can only have negative offsets - an offset of 0 shows the last line!
	if (nOffset > 0)
		nOffset = 0;

	g_nOutputBufOffset = nOffset;

	// Offset top by 
	srctReadRect.Top = (SHORT)(csbiInfo.dwSize.Y - (sc_nBotPaneLines - g_nOutputBufOffset));
	if (srctReadRect.Top < 0)
		srctReadRect.Top = 0;
	srctReadRect.Left = 0; 
	srctReadRect.Bottom = srctReadRect.Top + sc_nBotPaneLines;
	srctReadRect.Right = csbiInfo.dwSize.X - 1; 

	// The temporary buffer size is sc_nBotPaneLines rows x 80 columns. 
	coordBufSize.X = 80; 
	coordBufSize.Y = sc_nBotPaneLines; 

	// The top left destination cell of the temporary buffer is 
	// row 0, col 0. 
	coordBufCoord.X = 0; 
	coordBufCoord.Y = 0; 

	// Copy the block from the screen buffer to the temp. buffer. 

	bSuccess = ReadConsoleOutput( 
	   g_hOutputBuffer,        // screen buffer to read from 
	   chiBuffer,      // buffer to copy into 
	   coordBufSize,   // col-row size of chiBuffer 
	   coordBufCoord,  // top left dest. cell in chiBuffer 
	   &srctReadRect); // screen buffer source rectangle 
	if (!bSuccess) 
		return; 

	// Set the destination rectangle. 
	srctWriteRect.Top = sc_nBotPaneTextOffset+1; 
	srctWriteRect.Left = 0; 
	srctWriteRect.Bottom = (srctWriteRect.Top+sc_nBotPaneLines)-1; 
	srctWriteRect.Right = 80-1; 

	// Copy from the temporary buffer to the new screen buffer. 

	bSuccess = WriteConsoleOutput( 
		g_hStdOut, // screen buffer to write to 
		chiBuffer,        // buffer to copy from 
		coordBufSize,     // col-row size of chiBuffer 
		coordBufCoord,    // top left src cell in chiBuffer 
		&srctWriteRect);  // dest. screen buffer rectangle 
	if (!bSuccess) 
		return;

}


