/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


static void * WriteInvalid(DWORD dwAddress)
{
	CPU_Halt("Illegal Memory Access");

	DPF(DEBUG_MEMWARN, "Illegal Memory Access Tried to Write To 0x%08x PC: 0x%08x", dwAddress, g_dwPC);
	DBGConsole_Msg(0, "Illegal Memory Access: Tried to Write To 0x%08x (PC: 0x%08x)", dwAddress, g_dwPC);


	return MemoryBuffers[MEM_UNUSED];

}

static void *Write_Noise(DWORD dwAddress)
{
	static BOOL bWarned = FALSE;
	//if (!bWarned)
	{
	//	DBGConsole_Msg(0, "Writing noise (0x%08x) - sizing memory?", dwAddress);
		bWarned = TRUE;
	}
	return MemoryBuffers[MEM_UNUSED];

}

static void * WriteMappedGolden(DWORD dwAddress)
{
	//DWORD dwTranslated = dwGEOff;//(dwAddress- 0x7f000000) + GOLDENEYE_HACKED_OFFSET;
	//DWORD dwTranslated = dwAddress + GOLDENEYE_HACKED_OFFSET;
	//DBGConsole_Msg(0, "Writing Golden Mapped from 0x%08x, 0x%08x", dwAddress, dwTranslated);
	//return MemoryBuffers[MEM_UNUSED];


	return Write_Noise(dwAddress);

}


static void * WriteMapped(DWORD dwAddress)
{
//g_dwTLBWHit++;
	//DBGConsole_Msg(DEBUG_TLB, "Resolving Write Mapped Address");

	for (DWORD dwI = 0; dwI < 32; dwI++) {
		struct TLBEntry & tlb = g_TLBs[dwI];

		if ((dwAddress & ~tlb.mask) != (tlb.vpn2 & ~tlb.mask))
			continue;
		
		// Matched entry
		if (tlb.g)
		{
			// Global
			if (dwAddress & tlb.checkbit)
			{
				// Odd entry
				if (tlb.pfno & TLBLO_V)
				{
					// Valid
					DWORD dwAddr = 0x80000000 |
							       ((tlb.pfno<<TLBLO_PFNSHIFT) & ((~tlb.mask)>>1)) |
								   (dwAddress & (tlb.mask>>1));

					//DPF(DEBUG_TLB, "TLB Mapping: Write: 0x%08x -> 0x%08x (Index %d, odd)", dwAddress, dwAddr, dwI);
					return WriteAddress(dwAddr);
				}
				else
				{
					// Throw TLB Invalid exception
					DBGConsole_Msg(DEBUG_TLB, "TLB: Invalid Write");
					DPF(DEBUG_TLB, "Write: TLB Invalid Exception: 0x%08x (Index %d, odd)", dwAddress, dwI);
					return WriteInvalid(dwAddress);
				}
			} else {
				// Even entry 
				if (tlb.pfne & TLBLO_V) {
					// Valid
					DWORD dwAddr = 0x80000000 |
							       ((tlb.pfne<<TLBLO_PFNSHIFT) & ((~tlb.mask)>>1)) |
								   (dwAddress & (tlb.mask>>1));

					//DPF(DEBUG_TLB, "TLB Mapping: Write: 0x%08x -> 0x%08x (Index %d, even)", dwAddress, dwAddr, dwI);
					return WriteAddress(dwAddr);
				} else {
					// Throw TLB Invalid exception
					DBGConsole_Msg(DEBUG_TLB, "TLB: Invalid Write");
					DPF(DEBUG_TLB, "Write: TLB Invalid Exception: 0x%08x (Index %d, even)", dwAddress, dwI);
					return WriteInvalid(dwAddress);
				}
			}
		} else {
			// Not Global
			// Do ASID stuff
			DBGConsole_Msg(DEBUG_TLB, "TLB: ASID Write (NoImpl) (Index %d)", dwI);
			DPF(DEBUG_TLB, "TLB: ASID Write (NoImpl) (Index %d)", dwI);
			return WriteInvalid(dwAddress);
		}
	}
	// No valid TLB entry - throw TLB Miss Exception
	TLB_Refill_Store(dwAddress);
	return WriteInvalid(dwAddress);
}



static void *Write_RAM_4Mb_8000_803F(DWORD dwAddress)
{
	//DPF(DEBUG_MEMORY, "Writing to RDRAM: 0x%08x", dwAddress);
	return g_pu8RamBase + (dwAddress & 0x003FFFFF);
}

static void *Write_RAM_8Mb_8000_807F(DWORD dwAddress)
{
	//DPF(DEBUG_MEMORY, "Writing to RDRAM: 0x%08x", dwAddress);
	return g_pu8RamBase + (dwAddress & 0x007FFFFF);
}



// 0x03F0 0000 to 0x03FF FFFF  RDRAM registers
static void *Write_83F0_83F0(DWORD dwAddress)
{
	
	if ((dwAddress&0x1FFFFFFF) < 0x04000000)
	{
		DPF(DEBUG_MEMORY_RDRAM_REG, "Writing to MEM_RD_REG: 0x%08x", dwAddress);
	//	DBGConsole_Msg(0, "Writing to MEM_RD_REG: 0x%08x", dwAddress);

		return ((BYTE *)MemoryBuffers[MEM_RD_REG] + (dwAddress & 0xFF));
	}
	else
	{
		return WriteInvalid(dwAddress);
	} 
}

// 0x0400 0000 to 0x0400 FFFF  SP memory
static void * Write_8400_8400(DWORD dwAddress)
{

	if ((dwAddress&0x1FFFFFFF) <= SP_IMEM_END)
	{
		DPF(DEBUG_MEMORY_SP_IMEM, "Writing to SP_MEM: 0x%08x", dwAddress);

		return ((BYTE *)MemoryBuffers[MEM_SP_MEM] + (dwAddress & 0x1FFF));
	}
	else
	{	
		return WriteInvalid(dwAddress);
	}
}

