/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include <crtdbg.h>

#include <ddraw.h>
#include <d3dx.h>

#include "DBGConsole.h"

#ifndef SAFE_DELETE
#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#endif

#include "SurfaceHandler.h"
#include "GfxContext.h"


#ifdef LIBPNG_SUPPORT
#include <png.h>


// cexcept interface

static BOOL PngSaveImage (LPCTSTR pstrFileName, png_byte *pDiData,
                   int iWidth, int iHeight);

#endif
//////////////////////////////////////////
// Constructors / Deconstructors

SurfaceHandler::SurfaceHandler(DWORD dwWidth, DWORD dwHeight) :
	m_lpDDS(NULL),
	m_SurfFmt(D3DX_SF_UNKNOWN),
	m_dwWidth(0),
	m_dwHeight(0)

{
	Resize(dwWidth, dwHeight);
}


SurfaceHandler::~SurfaceHandler(void)
{

	if (m_lpDDS)
	{
		m_lpDDS->Release();
		m_lpDDS = NULL;
	}
	m_dwWidth = 0;
	m_dwHeight = 0;

}


static void Init()
{
	HRESULT hr;
	LPDIRECT3DDEVICE7 pD3DDev;
	DWORD dwFlags;
	DWORD dwWidth;
	DWORD dwHeight;
	D3DX_SURFACEFORMAT pf;

	g_GfxContext.Lock();
	pD3DDev = g_GfxContext.GetD3DDev();

	// Don't worry about Null pointer
	dwFlags = 0;
	dwWidth = 64;
	dwHeight = 64;
	pf = D3DX_SF_A8R8G8B8;
	hr = D3DXCheckTextureRequirements(pD3DDev, &dwFlags, &dwWidth, &dwHeight, &pf);
	if (FAILED(hr))
	{	
			
		
	}
	g_GfxContext.Unlock();

}

//////////////////////////////////////////
// Resize the DIB. Just passes control to
// CreateDIB()
BOOL
SurfaceHandler::Resize(DWORD dwWidth, DWORD dwHeight)
{
	LPDIRECTDRAWSURFACE7 lpSurf;

	if (dwWidth < 1) dwWidth = 1;
	else if (dwWidth > 1024) dwWidth = 1024;
	if (dwHeight < 1) dwHeight = 1;
	else if (dwHeight > 1024) dwHeight = 1024;

	// Check if surface is already of required dimensions etc
	if (m_lpDDS && dwWidth == m_dwWidth && dwHeight == m_dwHeight)
		return TRUE;

	if (dwWidth < m_dwWidth)
		DBGConsole_Msg(0, "New width (%d) < Old Width (%d)", dwWidth, m_dwWidth);
	if (dwHeight < m_dwHeight)
		DBGConsole_Msg(0, "New height (%d) < Old height (%d)", dwHeight, m_dwHeight);
	
	lpSurf = CreateSurface(dwWidth, dwHeight);
	if (lpSurf == NULL)
		return FALSE;

	// Copy from old surface to new surface
	if (m_lpDDS != NULL)
	{
		m_lpDDS->Release();
		m_lpDDS = NULL;
	}

	m_dwWidth = dwWidth;
	m_dwHeight = dwHeight;
	m_lpDDS = lpSurf;

	//GetSurfaceInfo();

	return TRUE;

}


/////////////////////////////////////////////
// Create the surface. Destroys old surface if
// necessary
LPDIRECTDRAWSURFACE7
SurfaceHandler::CreateSurface(DWORD dwWidth, DWORD dwHeight)
{
	HRESULT hr;
	LPDIRECTDRAWSURFACE7 lpSurf;
	LPDIRECT3DDEVICE7 pD3DDev;
	DWORD dwCorrectedWidth;
	DWORD dwCorrectedHeight;


	DWORD dwFlags = D3DX_TEXTURE_NOMIPMAP;		// D3DX_TEXTURE_STAGE0
	D3DX_SURFACEFORMAT pf = D3DX_SF_A8R8G8B8;
	DWORD dwNumMaps = 0;

	dwCorrectedWidth = dwWidth;
	dwCorrectedHeight = dwHeight;

	g_GfxContext.Lock();
		pD3DDev = g_GfxContext.GetD3DDev();
		hr = D3DXCreateTexture(pD3DDev, &dwFlags, &dwCorrectedWidth, &dwCorrectedHeight, &pf, NULL, &lpSurf, &dwNumMaps);
	g_GfxContext.Unlock();

	if (pf != D3DX_SF_A8R8G8B8 &&
		pf != D3DX_SF_A4R4G4B4)
	{
		static BOOL bWarned = FALSE;
		if (!bWarned)
		{
			MessageBox(g_hMainWindow, "Unsupported Texture format", g_szDaedalusName, MB_ICONEXCLAMATION|MB_OK);
			bWarned = TRUE;
		}
	}

	if (dwWidth != dwCorrectedWidth ||
		dwHeight != dwCorrectedHeight)
	{
		DBGConsole_Msg(0, "Couldn't create texture of size %d x %d (get %d x %d)",
			dwWidth, dwHeight, dwCorrectedWidth, dwCorrectedHeight);
		if (lpSurf)
			lpSurf->Release();
		return NULL;
	}

	// HACK - we should only assign this when m_lpDDS is assigned!
	m_SurfFmt = pf;
	

	if (FAILED(hr))
	{
		char szError[200+1];

		D3DXGetErrorString(hr, 200, szError);
		
		DBGConsole_Msg(0, "!!Unable to create surface!! %d x %d (%s)", dwWidth, dwHeight, szError);
		
		return NULL;
	}
	
	return lpSurf;		
}

HRESULT 
SurfaceHandler::GetSurfaceInfo()
{
	HRESULT hr;
	DDPIXELFORMAT pf;

	if (m_lpDDS == NULL)
		return E_FAIL;

	ZeroMemory(&pf, sizeof(DDPIXELFORMAT));
	pf.dwSize = sizeof(DDPIXELFORMAT);
	pf.dwFlags = (DDPF_ALPHAPIXELS|DDPF_RGB);
	hr = m_lpDDS->GetPixelFormat(&pf);
	if (FAILED(hr))
		return hr;

	//DBGConsole_Msg(0, "Pixel Format: bpp: %d", pf.dwRGBBitCount);
	//DBGConsole_Msg(0, "         Red: 0x%08x", pf.dwRBitMask);
	//DBGConsole_Msg(0, "       Green: 0x%08x", pf.dwGBitMask);
	//DBGConsole_Msg(0, "        Blue: 0x%08x", pf.dwBBitMask);
	//DBGConsole_Msg(0, "       Alpha: 0x%08x", pf.dwRGBAlphaBitMask);


	return S_OK;
}


HRESULT
SurfaceHandler::GetSurfaceFormat(D3DX_SURFACEFORMAT * pSurfFmt)
{
	if (m_lpDDS == NULL)
		return E_FAIL;
	
	*pSurfFmt = m_SurfFmt;

	return S_OK;
}

//////////////////////////////////////////////////
// Get information about the DIBitmap
// This locks the bitmap (and stops 
// it from being resized). Must be matched by a
// call to EndUpdate();
BOOL
SurfaceHandler::StartUpdate(DrawInfo *di, RECT *pRect)
{
	DDSURFACEDESC2 ddsd;
	
	if (m_lpDDS == NULL)
		return FALSE;

    ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
    ddsd.dwSize                 = sizeof(DDSURFACEDESC2);
    //ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);

	//TODO - Clip pRect?

	if (SUCCEEDED(m_lpDDS->Lock(pRect, &ddsd, DDLOCK_NOSYSLOCK|DDLOCK_WAIT , NULL)))
	{
		if (pRect)
		{
			CopyRect(&di->rcLockRect, pRect);
			di->bLockedEntireSurface = FALSE;
			//di->dwWidth = pRect->right - pRect->left;			// Is this right? Maybe use ddsd vals instead? 
			//di->dwHeight = pRect->bottom - pRect->top;
		} else {
			di->bLockedEntireSurface = TRUE;
		}
		di->dwWidth = ddsd.dwWidth;
		di->dwHeight = ddsd.dwHeight;
		di->lpSurface = ddsd.lpSurface;
		di->lPitch    = ddsd.lPitch;
	} else {
		return FALSE;
	}

	return TRUE;
}

///////////////////////////////////////////////////
// This releases the DIB information, allowing it
// to be resized again
void
SurfaceHandler::EndUpdate(DrawInfo *di)
{
	RECT *pRC;
	
	if (m_lpDDS == NULL)
		return;

	// Check result of Unlock?
	if (di->bLockedEntireSurface) 
		pRC = &di->rcLockRect;
	else
		pRC = NULL;

	m_lpDDS->Unlock(pRC);

}



//////////////////////////////////////////////////////
// Set all pixels to (0,0,0) (also clears alpha)
void
SurfaceHandler::Clear(void)
{
	DDBLTFX ddbltfx;
	
	if (m_lpDDS == NULL)
		return;

	ZeroMemory(&ddbltfx, sizeof(ddbltfx));
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = 0;

	m_lpDDS->Blt(NULL, NULL, NULL, DDBLT_COLORFILL|DDBLT_WAIT, &ddbltfx);

}


// Make the pixels of the texture blue
HRESULT 
SurfaceHandler::MakeBlue(DWORD dwWidth, DWORD dwHeight)
{
	DrawInfo di;
	
	if (!StartUpdate(&di, NULL))
		return E_FAIL;
	
	DWORD * pSrc = (DWORD *)di.lpSurface;
	LONG lPitch = di.lPitch;

	for (DWORD y = 0; y < dwHeight; y++)
	{
		DWORD * dwSrc = (DWORD *)((BYTE *)pSrc + y*lPitch);
		for (DWORD x = 0; x < dwWidth; x++)
		{
			dwSrc[x] = 0x000000FF | ((rand()% 255)<<8) | ((rand()%255)<<24);
		}
	}
	
	EndUpdate(&di);

	return S_OK;

}


BOOL SurfaceHandler::DumpImageAsRAW(LPCTSTR szFileName, DWORD dwWidth, DWORD dwHeight)
{
	DrawInfo di;
	FILE * fp = fopen(szFileName, "wb");
	if (fp == NULL)
		return FALSE;
	
	if (!StartUpdate(&di, NULL))
	{
		fclose(fp);
		return FALSE;
	}
	
	DWORD *pSrc = (DWORD *)di.lpSurface;
	LONG lPitch = di.lPitch;

	for (DWORD y = 0; y < dwHeight; y++)
	{
		DWORD * dwSrc = (DWORD *)((BYTE *)pSrc + y*lPitch);
		for (DWORD x = 0; x < dwWidth; x++)
		{
			DWORD dw = dwSrc[x];

			BYTE dwRed   = (BYTE)((dw & 0x00FF0000)>>16);
			BYTE dwGreen = (BYTE)((dw & 0x0000FF00)>>8 );
			BYTE dwBlue  = (BYTE)((dw & 0x000000FF)    );
			BYTE dwAlpha = (BYTE)((dw & 0xFF000000)>>24);

			fwrite(&dwRed, 1,1,fp);
			fwrite(&dwGreen, 1,1,fp);
			fwrite(&dwBlue, 1,1,fp);
			fwrite(&dwAlpha, 1,1,fp);
		}
	}
	EndUpdate(&di);

	fclose(fp);

	return TRUE;
}

#ifdef LIBPNG_SUPPORT

BOOL SurfaceHandler::DumpImageAsPNG(LPCTSTR szFileName, DWORD dwWidth, DWORD dwHeight)
{
	DWORD * pB = new DWORD [dwWidth * dwHeight];
	if (pB == NULL)
	{
		return FALSE;
	}

	DrawInfo di;
	
	if (!StartUpdate(&di, NULL))
	{
		delete [] pB;
		return FALSE;
	}
	
	DWORD *pSrc = (DWORD *)di.lpSurface;
	LONG lPitch = di.lPitch;

	DWORD index = 0;

	for (DWORD y = 0; y < dwHeight; y++)
	{
		DWORD * dwSrc = (DWORD *)((BYTE *)pSrc + y*lPitch);
		for (DWORD x = 0; x < dwWidth; x++)
		{
			pB[index++] = dwSrc[x];
		}
	}
	EndUpdate(&di);
	
	PngSaveImage(szFileName, (png_byte *)pB, dwWidth, dwHeight);

	delete [] pB;

	return TRUE;
}



static void
png_cexcept_error(png_structp png_ptr, png_const_charp msg)
{
	// Do nothing
}

static png_structp png_ptr = NULL;
static png_infop info_ptr = NULL;


BOOL PngSaveImage (LPCTSTR pstrFileName, png_byte *pDiData,
                   int iWidth, int iHeight)
{
    const int           ciBitDepth = 8;
    const int           ciChannels = 4;

    static FILE        *pfFile;
    png_uint_32         ulRowBytes;
    static png_byte   **ppbRowPointers = NULL;
    int                 i;

    // open the PNG output file

    if (!pstrFileName)
        return FALSE;

    if (!(pfFile = fopen(pstrFileName, "wb")))
        return FALSE;

    // prepare the standard PNG structures

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL,
      (png_error_ptr)png_cexcept_error, (png_error_ptr)NULL);
    if (!png_ptr)
    {
        fclose(pfFile);
        return FALSE;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
	{
        fclose(pfFile);
        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
        return FALSE;
    }

    //Try
    {
        // initialize the png structure
		png_init_io(png_ptr, pfFile);
    
        // we're going to write a very simple 3x8 bit RGB image
        
        png_set_IHDR(png_ptr, info_ptr, iWidth, iHeight, ciBitDepth,
            PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
            PNG_FILTER_TYPE_BASE);
        
        // write the file header information
        
        png_write_info(png_ptr, info_ptr);
        
        // swap the BGR pixels in the DiData structure to RGB
        
        png_set_bgr(png_ptr);
        
        // row_bytes is the width x number of channels
        
        ulRowBytes = iWidth * ciChannels;
        
        // we can allocate memory for an array of row-pointers
        
        if ((ppbRowPointers = (png_bytepp) malloc(iHeight * sizeof(png_bytep))) == NULL)
		{
            //Throw "Visualpng: Out of memory";
		}
        
        // set the individual row-pointers to point at the correct offsets
        
        for (i = 0; i < iHeight; i++)
            ppbRowPointers[i] = pDiData + i * (((ulRowBytes + 3) >> 2) << 2);
        
        // write out the entire image data in one call
        
        png_write_image (png_ptr, ppbRowPointers);
        
        // write the additional chunks to the PNG file (not really needed)
        
        png_write_end(png_ptr, info_ptr);
        
        // and we're done
        
        free (ppbRowPointers);
        ppbRowPointers = NULL;
        
        // clean up after the write, and free any memory allocated
        
        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
        
        // yepp, done
    }

    fclose (pfFile);
    
    return TRUE;
}
#endif // LIBPNG_SUPPORT
